"""
    Redis functions
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import hashlib
import logging
import time

from redis import StrictRedis

logger = logging.getLogger(__name__)


class RedisHandler:
    """
        Handler For redis database
    """

    def __init__(self, url, retry_count=3, retry_delay=2):
        """

        Args:
            url (str): redis server url
            retry_count (int) : how many retry in error condition
            retry_delay (int0 :
        """
        self.url = url
        self.redis_cache = StrictRedis.from_url(self.url, decode_responses=True)
        self.retry_count = retry_count
        self.retry_delay = retry_delay

    def _get(self, key: str):
        try:
            return self.redis_cache.get(key)
        except Exception as e:
            logger.error(f"Error in get from cache e: {e}")
            return None

    def get(self, key, default=None, cast=None) -> str:
        """
            get value in redis
        Args:
            key (str): name of value in database
            default: default value
            cast: cast function
        Returns:
            (str) : value in redis

        """
        for i in range(self.retry_count):
            value = self._get(key=key)
            if value is None:
                time.sleep(self.retry_delay)
                continue
            if value is None or value == "None":
                logger.warning(f"{key} not found in cache. initialise it.")
                self.set(key, default)
            try:
                return cast(self._get(value))
            except Exception as e:
                logger.error(f"Error in get from cache e: {e}")
                return "None"
        return "None"

    def set(self, key, value) -> bool:
        """
            Set a value in redis
        Args:
            key (str): name of value
            value (str): value

        Returns:
            (bool) : str
        """
        value = str(value)
        for i in range(self.retry_count):
            try:
                self.redis_cache.set(key, value)
                return True
            except Exception as e:
                logger.error(f"Error in set in cache. e: {e}")
                time.sleep(self.retry_delay)
                continue
        return False

    def set_hash(self, key, data) -> bool:
        """
            Set hash of value in redis
        Args:
            key (str): name of key
            data (str): data
        Returns:
            (bool) : result
        """
        return self.set(key, self.get_hash(data))

    @staticmethod
    def get_hash(data) -> str:
        """
            Calculate hash of data

        Args:
            data : input data

        Returns:
            (str) : hash value
        """
        return hashlib.md5(str(data).encode()).hexdigest()

    def check_hash(self, key, data=None) -> bool:
        """
            Check hash of data by saved data
        Args:
            key (str): input key
            data: input data

        Returns:
            (bool) : result of checking
        """
        if data is None:
            return False
        saved_hash = self.get(key=key)
        return self.get_hash(data) == saved_hash



"""
    Easy tools for nice string and printing
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import logging
import os
from datetime import datetime
from enum import Enum
from typing import Tuple, List

try:
    from blessings import Terminal
except ImportError:
    Terminal = None

logger = logging.getLogger(__name__)


class Colors(Enum):
    """
        All Color for beautiful result
    """
    DEFAULT = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    NO_UNDERLINE = '\033[24m'
    NEGATIVE = '\033[7m'
    POSITIVE = '\033[27m'
    BLACK_F = '\033[30m'
    RED_F = '\033[31m'
    GREEN_F = '\033[32m'
    YELLOW_F = '\033[33m'
    BLUE_F = '\033[34m'
    MAGENTA_F = '\033[35m'
    CYAN_F = '\033[36m'
    WHITE_F = '\033[37m'
    EXTENDED_F = '\033[38m'
    DEFAULT_F = '\033[39m'
    BLACK_B = '\033[40m'
    RED_B = '\033[41m'
    GREEN_B = '\033[42m'
    YELLOW_B = '\033[43m'
    BLUE_B = '\033[44m'
    MAGENTA_B = '\033[45m'
    CYAN_B = '\033[46m'
    WHITE_B = '\033[47m'
    EXTENDED_B = '\033[48m'
    DEFAULT_B = '\033[49m'
    BRIGHT_BLACK_F = '\033[90m'
    BRIGHT_RED_F = '\033[91m'
    BRIGHT_GREEN_F = '\033[92m'
    BRIGHT_YELLOW_F = '\033[93m'
    BRIGHT_BLUE_F = '\033[94m'
    BRIGHT_MAGENTA_F = '\033[95m'
    BRIGHT_CYAN_F = '\033[96m'
    BRIGHT_WHITE_F = '\033[97m'
    BRIGHT_BLACK_B = '\033[100m'
    BRIGHT_RED_B = '\033[101m'
    BRIGHT_GREEN_B = '\033[102m'
    BRIGHT_YELLOW_B = '\033[103m'
    BRIGHT_BLUE_B = '\033[104m'
    BRIGHT_MAGENTA_B = '\033[105m'
    BRIGHT_CYAN_B = '\033[106m'
    BRIGHT_WHITE_B = '\033[107m'


def get_colors(color=None, colors=None):
    """
        get color list
    Args:
        color (Colors): 
        colors (List[Colors]): 

    Returns:
        (str) : color list by string

    """
    result = ""
    if color is not None:
        result += color.value
    if colors is not None:
        for color in colors:
            result += color.value
    return result


def get_text(*args, sep=' '):
    """
        Use this function for join some data with a separator
    Args:
        *args: all arguments
        sep (str): text separator

    Returns:
        (str) : result text

    """
    text = ""
    for i in range(len(args)):
        text += str(args[i])
        if i != len(args) - 1:
            text += sep
    return text


def get_color_text(*args, color=None, colors=None, sep=' '):
    """
        Use this function for join some data with a separator and return colorful string
    Args:
        *args: all input arguments
        color (Colors): text color
        colors (List[Colors]): text colors
        sep (str): text separator

    Returns:
        (str) : result colorful text

    """
    text = get_text(*args, sep=sep)
    return f"{get_colors(color=color, colors=colors)}{text}{Colors.DEFAULT.value}"


def print_color(*args, end: str = '\n', color=None, colors=None, sep=' '):
    """
        Print colorful text
    Args:
        *args: all input arguments
        end (str): end text
        color (Colors): text color
        colors (List[Colors]): text colors 
        sep (str): text separator 

    Returns:
        (str) : printed text

    """
    printed_text = f"{get_color_text(*args, sep=sep, colors=colors, color=color)}"
    print(printed_text, end=end)
    return printed_text


class Progressbar:
    """
        Use this class for show progressbar in loop or etc.
    """

    def __init__(self, prefix="", suffix="", fill='█', not_fill='-', left_schema=" |", right_schema="| ", decimals=1,
                 show_time=False):
        """
            Progressbar initialize data
        Args:
            prefix (str): text shown before progressbar
            suffix (str): text shown after progressbar
            fill (str): a character shown for filling progressbar  
            not_fill (str): a character shown for not filled in progressbar
            left_schema (str): a character use for left of filling string 
            right_schema (str): a character use for right of not filling string 
            decimals (int): decimal point for completed percent
            show_time (bool): for showing elapsed and remaining time
        """
        self.prefix = prefix
        self.suffix = suffix
        self.fill = fill
        self.not_fill = not_fill
        self.left_schema = left_schema
        self.right_schema = right_schema
        self.decimals = decimals
        self.elapsed_time = ""
        self.remain_time = ""

        self.prefix_color = Colors.BRIGHT_MAGENTA_F
        self.suffix_color = Colors.MAGENTA_F
        self.fill_color = Colors.BLUE_F
        self.not_fill_color = Colors.YELLOW_F
        self.left_schema_color = Colors.RED_F
        self.right_schema_color = Colors.RED_F
        self.percent_color = Colors.CYAN_F
        self.time_color = Colors.GREEN_F

        self.is_print = False
        self.terminal = None
        self.show_time = show_time
        self.start_time = None
        self._set_start_time()
        if Terminal is not None:
            try:
                self.terminal = Terminal()
            except Exception as e:
                logger.error(f"You are not in terminal {e}")

    def _set_start_time(self):
        self.start_time = datetime.now()

    @property
    def _colored_prefix(self):
        return get_color_text(self.prefix, color=self.prefix_color)

    @property
    def _colored_suffix(self):
        return get_color_text(self.suffix, color=self.suffix_color)

    @property
    def _colored_fill(self):
        return get_color_text(self.fill, color=self.fill_color)

    @property
    def _colored_not_fill(self):
        return get_color_text(self.not_fill, color=self.not_fill_color)

    @property
    def _colored_left_schema(self):
        return get_color_text(self.left_schema, color=self.left_schema_color)

    @property
    def _colored_right_schema(self):
        return get_color_text(self.right_schema, color=self.right_schema_color)

    @property
    def _colored_elapsed_time(self):
        return get_color_text(self.elapsed_time, color=self.time_color)

    @property
    def _colored_remain_time(self):
        return get_color_text(self.remain_time, color=self.time_color)

    def _get_percent(self, iteration: int, total: int) -> Tuple[str, int]:
        elapsed_time = (datetime.now() - self.start_time).total_seconds()
        self.elapsed_time = ""
        self.remain_time = ""
        if self.show_time:
            self.elapsed_time = f"elapsed {int(elapsed_time)} "
            if iteration > 0:
                remain_time = (elapsed_time * (total - iteration) / iteration) + 1
                self.remain_time = f" remain {int(remain_time)}"

        percent = ("{0:." + str(self.decimals) + "f}").format(100 * (iteration / float(total))) + "% "
        minus_len = (len(self.prefix) + len(self.left_schema) + len(self.right_schema) + len(self.suffix) +
                     len(self.remain_time) + len(self.elapsed_time))
        length = self._terminal_size - len(percent) - minus_len
        return get_color_text(percent, color=self.percent_color), length

    # noinspection PyBroadException
    @property
    def _terminal_size(self):
        """
        update terminal size
        :return:
        """
        try:
            if self.terminal is None:
                terminal_size = os.get_terminal_size().columns
            else:
                terminal_size = self.terminal.width
            if terminal_size is None:
                raise TypeError
        except Exception:
            return -1
        return terminal_size

    def print(self, iteration, total):
        """
            Print progressbar
            if total >= iteration go to next line
        Args:
            iteration (int): current position of progressbar
            total (int): total data count for progressbar

        Returns:


        """
        if iteration <= 1:
            self._set_start_time()
        if self._terminal_size == -1:
            print("-", end="")
        else:
            percent, length = self._get_percent(iteration=iteration, total=total)
            filled_length = int(length * iteration // total)
            print(
                self._colored_elapsed_time,
                self._colored_prefix,
                self._colored_left_schema,
                self._colored_fill * filled_length,
                self._colored_not_fill * (length - filled_length),
                self._colored_right_schema,
                percent,
                self._colored_suffix,
                self._colored_remain_time,
                sep="", end="\r")

        # Print New Line on Complete
        if iteration <= total:
            print()



"""
    Worker Thread
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import logging
import multiprocessing
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import datetime
from queue import Queue
from threading import Thread, Lock

from func_timeout import func_timeout
from func_timeout.exceptions import FunctionTimedOut

logger = logging.getLogger(__name__)


class Worker(Thread):
    """
        Thread executing tasks from a given tasks queue
    """

    def __init__(self, pool: "ThreadPool", index=None):
        name = f"{pool.name} {index}"
        Thread.__init__(self, name=name)
        self.name = name
        self.pool = pool
        self.is_working = False
        self.daemon = True
        self.start()

    def run(self):
        """
            Run
        """
        while True:
            task_detail, func, args, kwargs = self.pool.tasks.get()
            task_id = task_detail["id"]
            timeout = None
            if type(func) == tuple:
                func, timeout = func
            task_detail["start_dt"] = datetime.now()
            self.is_working = True
            if not callable(func):
                task_detail["status"] = "not_run"
                task_detail["end_dt"] = datetime.now()
                logger.info(f"Task {task_id} is not callable in {self.name}")
                continue
            try:
                task_detail["status"] = "running"
                logger.debug(f"Start running function {func.__name__} in {self.name}.")
                result = func_timeout(timeout, func=func, args=args, kwargs=kwargs)
                task_detail["status"] = "done"
                task_detail["result"] = result
            except FunctionTimedOut:
                logger.error(
                    f"Timeout in function {func.__name__} in task {task_id} thread {self.name} after {timeout}")
                task_detail["status"] = "timeout"
            except Exception as e:
                logger.error(f"Error in function {func.__name__} in task {task_id} thread {self.name}. e: {e}")
                task_detail["status"] = "error"
            finally:
                self.pool.tasks.task_done()
                logger.debug(
                    f"Finish running function {func.__name__} in thread {self.name}.", func.__name__, self.name)
                task_detail["end_dt"] = datetime.now()
                duration = task_detail["end_dt"] - task_detail["insert_dt"]
                if self.pool.log_detail:
                    logger.info(f"Task {task_id} Complete in {self.name} in {duration}", task_detail)
            self.is_working = False


class ThreadPool:
    """
        Pool of threads consuming tasks from a queue
    """

    def __init__(self, worker_count: int = 4, total_queue: int = 20, *, name: str = __name__,
                 save_detail: bool = False, log_detail: bool = False):
        self.tasks = Queue(total_queue)
        self.name = name
        self.tasks_data = {}
        self.save_detail = save_detail
        self.log_detail = log_detail
        self.tasks_mutex = Lock()
        logger.info(f"{self.name} Create {worker_count} Worker With MAX_QUEUE: {total_queue}")
        self.workers = []
        for index in range(worker_count):
            self.workers.append(Worker(self, index=index))

    def add_task(self, func, *args, **kargs) -> int:
        """Add a task to the queue"""
        self.tasks_mutex.acquire()
        task_id = len(self.tasks_data.keys()) + 1
        task_detail = {
            "id": task_id,
            "status": "pending",
            "insert_dt": datetime.now()
        }
        if self.save_detail:
            self.tasks_data[task_id] = task_detail

        self.tasks.put((task_detail, func, args, kargs))
        self.tasks_mutex.release()
        return task_id

    def wait_completion(self):
        """
            Wait for completion of all the tasks in the queue
        """
        self.tasks.join()

    def get_task_data(self, task_id: int) -> dict:
        """
            Get task data
        """
        return self.tasks_data.get(task_id)

    def get_free_worker(self):
        """
            for get number of free worker
        """
        free_worker = 0
        for worker in self.workers:
            if not worker.is_working:
                free_worker += 1
        return free_worker


class ProcessThreadPool:
    """
        Combine Process and thread
    """

    def __init__(self, max_process: int = 8, max_thread: int = 20):
        self.max_process = max_process
        self.max_thread = max_thread
        self.handler = None

    def process_run(self, iterable: list, handler, job_per_process: int = None):
        """
            for run process
        """
        if not callable(handler):
            return
        self.handler = handler
        process_count = self.max_process
        chunk_size = (len(iterable) // process_count) + 1
        if job_per_process is not None:
            process_count = min(len(iterable) // job_per_process, self.max_process)
            process_count = max(process_count, 1)
            chunk_size = job_per_process

        chunk_path_list: [list] = []
        index = 0
        while index < len(iterable):
            chunk_path_list.append(iterable[index:index + chunk_size])
            index += chunk_size

        pool = multiprocessing.Pool(process_count)
        logger.info(f"start make {process_count} process")
        pool.map(func=self.thread_run, iterable=chunk_path_list)
        pool.close()
        pool.join()

    def thread_run(self, iterable: list, handler=None):
        """
            for run thread
        """
        if handler is None:
            handler = self.handler
        if not callable(handler):
            return
        with ThreadPoolExecutor(max_workers=min(len(iterable), self.max_thread)) as executor:
            executor.map(handler, iterable)



"""
Setup auto_utils
"""
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

packages = [
    'autoutils',
    # 'matrixcli'
]

install_requires = [
    "persiantools~=2.0",
    "redis~=4.0",
    'requests~=2.25',
    'urllib3~=1.26',
    "sshtunnel~=0.4",
    "paramiko~=2.7",
    "func-timeout~=4.3",
    "blessings~=1.7"
]

setup(
    name="autoutils",
    version="0.0.3",
    author="Reza Zeiny",
    author_email="rezazeiny1998@gmail.com",
    description="Some Good Function",
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rezazeiny/autoutils",
    packages=packages,
    platforms=['linux'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
    All Work with linux scripting
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import logging
import random
import string
import subprocess

from .color import get_text

logger = logging.getLogger(__name__)

PIPE = -1
STDOUT = -2
DEVNULL = -3


def id_generator(size: int = 6, chars: list = string.ascii_uppercase + string.ascii_lowercase + string.digits) -> str:
    """
        Generate random string
    Args:
        size (int): size of random string
        chars (list): define select set

    Returns:
        (str) : random string
    """
    return ''.join(random.choice(chars) for _ in range(size))


class RunScriptResult:
    """
        Object for run_script result
    """

    def __init__(self):
        self.command = None
        self.process_id = None
        self.exit_code = None
        self.stdout = None
        self.stderr = None
        self.error = None


def run_script(*args, timeout=None, background=False, quiet=False) -> "RunScriptResult":
    """
        Run Shell script in linux
    Args:
        *args: input command
        timeout (int): timeout for running command
        background (bool): run command in background
        quiet (bool): run command without any result
    Returns:
        (RunScriptResult) : result object
    """
    """
    :param save_output: for return output
    :param quiet: for run in quiet mode
    :param timeout: timeout
    :param background: background run
    :return: output_code output_content
    """
    result = RunScriptResult()
    result.command = get_text(*args)
    if background:
        result.command += " &"
    if quiet:
        stderr = DEVNULL
        stdout = DEVNULL
    else:
        stdout = PIPE
        stderr = PIPE
    logger.debug(f"run [{result.command}]: Start")

    try:
        with subprocess.Popen(result.command, shell=True, stdout=stdout, stderr=stderr) as process:
            result.process_id = process.pid
            stdout, stderr = "", ""
            try:
                if background:
                    process.wait(timeout=timeout)
                else:
                    stdout, stderr = process.communicate(timeout=timeout)
            # except subprocess.TimeoutExpired as e:
            #     process.kill()
            except Exception as e:  # Including KeyboardInterrupt, TimeoutExpired, communicate handled that.
                process.kill()
                raise e
            result.exit_code = process.poll()
            if result.exit_code == 0:
                logger.debug(f"run [{result.command}]: Complete")
            else:
                logger.debug(f"run [{result.command}]: Error Code: {result.exit_code}")
            if stdout is not None:
                if type(stdout) == bytes:
                    result.stdout = stdout.decode().split("\n")[:-1]
                else:
                    result.stdout = []
            if stderr is not None:
                if type(stderr) == bytes:
                    result.stderr = stderr.decode().split("\n")[:-1]
                else:
                    result.stderr = []
    except Exception as e:
        logger.error(f"error in run_script. e: {e}")
        result.error = e
    return result
